/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

/**
 * @test TestDynShrinkHeap
 * @key gc
 * @bug 8037924
 * @summary Verify that heap shrinking after full gc and dynamically changed free ratio
 * @author Andrey Zakharov /andrey.x.zakharov@oracle.com/
 * @library /testlibrary
 * @run main/othervm -XX:+UseAdaptiveSizePolicyWithSystemGC -XX:+UseParallelGC -XX:MinHeapFreeRatio=1
 * -XX:MaxHeapFreeRatio=100 -Xloggc:gc.log TestDynShrinkHeap
 */
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryUsage;
import java.util.ArrayList;
import com.oracle.java.testlibrary.TestDynamicVMOption;
import com.oracle.java.testlibrary.MemoryUsagePrinter;

public class TestDynShrinkHeap {

    public static final String MinFreeRatioFlagName = "MinHeapFreeRatio";
    public static final String MaxFreeRatioFlagName = "MaxHeapFreeRatio";

    public static ArrayList<byte[]> list = new ArrayList<byte[]>(0);
    public static byte[] temp;
    public static final int M = 1024 * 1024; // to make heap more manageable by test code

    private final TestDynamicVMOption minRatioOption;
    private final TestDynamicVMOption maxRatioOption;
    
    public final class Labels {
        public static final String beforeAllocation = "before allocation";
        public static final String arrayHalfed      = "array halfed";
        public static final String underPressure    = "under pressure";
        public static final String beforeGc         = "before full GC";
        public static final String afterGc          = "after full GC";
        
    }
    

    public TestDynShrinkHeap() {
        minRatioOption = new TestDynamicVMOption(MinFreeRatioFlagName);
        maxRatioOption = new TestDynamicVMOption(MaxFreeRatioFlagName);
    }

    public void test() {
        int minHeapFreeValue = minRatioOption.getIntValue();
        int maxHeapFreeValue = maxRatioOption.getIntValue();

        // with maxHeapFreeRatio = 100 heap will not shrink at all, so
        if (maxHeapFreeValue == 100) {
            maxHeapFreeValue = 95;
            maxRatioOption.setIntValue(maxHeapFreeValue);
        }

        MemoryUsagePrinter.printMemoryUsage(Labels.beforeAllocation);

        for (int i = 0; i < M; i++) {
            list.add(new byte[1024]);
        }

        MemoryUsagePrinter.printMemoryUsage("allocated " + M + " arrays");

        list.subList(0, M / 2).clear();
        gc();
        MemoryUsagePrinter.printMemoryUsage(Labels.arrayHalfed);
        MemoryUsage muFull = ManagementFactory.getMemoryMXBean().getHeapMemoryUsage();

        maxRatioOption.setIntValue(minHeapFreeValue + 1);
        gc();
        MemoryUsagePrinter.printMemoryUsage(Labels.underPressure);
        MemoryUsage muPressure = ManagementFactory.getMemoryMXBean().getHeapMemoryUsage();

        if (!(muPressure.getCommitted() < muFull.getCommitted())) {
            StringBuilder strb = new StringBuilder("committed heap size under pressure is not less than committed full heap size, heap hasn't been shrunk?");
            strb.append(System.getProperty("line.separator"));
            strb.append(MinFreeRatioFlagName + " = " + minHeapFreeValue);
            strb.append(System.getProperty("line.separator"));
            strb.append(MaxFreeRatioFlagName + " = " + maxHeapFreeValue);
            throw new RuntimeException(strb.toString());
        }
    }

    public static void main(String[] args) {
        new TestDynShrinkHeap().test();
    }

    public static void gc() {
        MemoryUsagePrinter.printMemoryUsage(Labels.beforeGc);
        System.gc();
        MemoryUsagePrinter.printMemoryUsage(Labels.afterGc);
    }

}
